<div>
	<sly-editor
		v-model="state.{{{widget.widget_id}}}.text"
		:options="data.{{{widget.widget_id}}}.editor_options"
	>
		<template slot="controls">
			{% if widget._restore_button is not none %} {{{ widget._restore_button }}}
			{% endif %}</template
		>
	</sly-editor>
</div>
